/* address_hash.c: generate/parse spam protected news addresses */
/*
    Copyright (C) 1998 Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@ohse.de, Uwe Ohse @ DU3 (mausnet)
*/
#include "tomsysdep.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <portab.h>
#include <errno.h>
#include <unistd.h>
#include <ctype.h>
#include <unistd.h>
#include <fcntl.h>
#include "str_ulong.h"

#include "box_conf.h"
#include "prototyp.h" /* entumlauten */
#include "user.h"
#include "mtzeit.h"
#include "address_hash.h"
#include "crc32.h"

/*
 * Verfahren:
 * So der Absender es wnscht wird die Absenderadresse ausgehender
 * ffentliche Nachrichten wie folgt verschlsselt:
 * UNR-expire-MTDATUM-HASH@box
 *     ^string
 * expire ist "Sekunden seit 1970", im MausTauschformat.
 * HASH wird so erzeugt:
 *   crc32(string(hash_secret).string(usernr).string(expire)).
 *   im base10 format abgelegt.
 * Bei Bedarf kann man das leicht auf md5 erweitern, aber md5 hat
 * den deutlichen Nachteil 128 Bit zu brauchen - das sind immerhin
 * 26 Zeichen bei base36 Kodierung.
 * Notwendigkeit zur cryptographischen Verschlsselung besteht 
 * sowieso nicht, es soll nur kein Spammer herausbekommen knnen
 * wie man korrekte Adressen erzeugt ...
 *
 * Beim umgekehrten Weg wird aus der Usernummer, dem privaten
 * hash_secret und dem bekannten Expire ein hash erzeugt und
 * mit der Adresse verglichen.
 */

char *
address_hash_create(USER *u,time_t expire)
{
	char numbuf[STR_ULONG*3];
	ULONG crc;
	int l;

	l=str_ulong(numbuf,u->address_hash_secret);
	numbuf[l++]='.';
	l+=str_ulong(numbuf+l,u->nummer);
	numbuf[l++]='.';
	l+=str_ulong(numbuf+l,expire);
	crc=crc32ansi(numbuf,l);

	/* now generate the localpart */
	l=str_ulong(numbuf,u->nummer);
	memcpy(numbuf+l,"-expire-",8);
	l+=8;
	time_t_to_mtzeit_sec(&expire,numbuf+l);
	l=strlen(numbuf);
	numbuf[l++]='-';
	l+=str_ulong(numbuf+l,crc);
	return XSTRDUP(numbuf);
}

int
address_hash_parse(const char *s, USERNR *unr)
{
	char *p;
	char *q;
	time_t expire;
	USER u;
	char *vgl;
	*unr=strtol(s,&p,10);
	if (*unr<1)
		return 0; /* not our business */
	if (memcmp(p,"-expire-",8))
		return 0; /* not our business */
	p+=8;
	q=strchr(p,'-');
	if (!q)
		return -1; /* crippled */
	if (q!=p+14)
		return -1; /* crippled */
	*q=0; /* modify const string ... */
	if (!mtzeit_to_time_t(p,&expire)) {
		*q='-';
		return -2; /* illegal date */
	}
	*q='-';
	if (!ub_read_user(*unr,&u,FALSE)) {
		return -3; /* unknown user */
	}
	vgl=address_hash_create(&u,expire);
	if (0!=strcasecmp(vgl,s)) {
		return -4; /* Flschung */
	}
	if (time(NULL)>expire)
		return -5;
	return 1;
}

ULONG
address_hash_gen_key(void)
{
#ifdef linux
	/* using /dev/urandom is a bit of overkill here ... */
	/* especially since this whole stuff is overkill :-) */
	int fd=open("/dev/urandom",O_RDONLY);
	if (fd) {
		ULONG r=0;
		while (r==0) {
			if (sizeof(r)==read(fd,&r,sizeof(r))) {
				if (r) {
					close(fd);
					return r;
				}
			}
		}
	}
#endif
	{
		static BOOLEAN was_here_before=FALSE;
		ULONG r=0;
		if (!was_here_before) {
			was_here_before=TRUE;
			srand(time(0)); /* best entropy i have */
		}
		while (!r) {
			/*
			 * hope RAND_MAX is _large_
			 */
			r=rand();
		}
		return r;
	}
}
