/* tombrett.c: utility program for the brett library.
    Copyright (C) 1996 Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
       Snail Mail (don't expect me to answer):
             Uwe Ohse
             Drosselstrae 2
             47055 Duisburg
             Germany
*/
#include "tomsysdep.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <getopt.h>
#include <regex.h>
#include <fnmatch.h>

#include "libquark.h"
#include "brett_db.h"

#include <limits.h>
#include <ctype.h>
#include <time.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <syslog.h>
#include <errno.h>

#include "long-options.h"
#include "tinterp.h" /* damit sie in . liegen darf */
const char *program_name;

static size_t my_tip_percent_fn(tip_t *tab, const char *this, 
	const char **next, char *to, size_t maxlen);

/* textinterpretationstabelle */
static tip_t tiptab[]=
{
	{ '\\', NULL,        0, tip_backslash},
	{ '~',  NULL,        0, tip_tilde},
	{ '%',  NULL,        0, my_tip_percent_fn},
	{ '\0', "$[",        0, tip_conditional},
	{ '\0', "${strip",   0, tip_strip},
	{ '$',  NULL,        0, tip_dollar},
	{ '\0', NULL,        0, NULL}
};

BRETT *akt_brett;

/* ========================  Utility - Routinen  ======================== */

#define ADDC(x) do {\
	if (2>=maxlen) { tip_errtext="buffer shortage"; return TIP_ERROR;}; \
	if (x) \
		*to++=x; \
	*to='\0'; \
	maxlen--; \
	} while(0)

static size_t 
my_tip_percent_fn(tip_t *tab TOM_ATTRIB_UNUSED, const char *this, const char **next,
    char *to, size_t maxlen)
{
	size_t l;
	char buf[512]="";
	size_t tmaxlen=-1;
	size_t tfixlen=-1;
	char pad=' ';
	size_t retl;

	this++; /* skip over % */
	if (isdigit(*this))
	{
		char *nextchar;
		int old=errno;
		errno=0;
		tmaxlen=strtoul(this,&nextchar,10);
		if (errno)
			return TIP_NOTDONE;
		errno=old;
		if (*this=='0')
			pad='0';
		this=nextchar;
	}
	if (*this=='.')
	{
		char *nextchar;
		int old=errno;
		this++; /* skip over dot */
		errno=0;
		tfixlen=strtoul(this,&nextchar,10);
		if (errno)
			return TIP_NOTDONE;
		errno=old;
		this=nextchar;
	}
	if (!*this)
		return TIP_NOTDONE;

	*next=this+1;
	switch(*this)
	{
	case '%':
		strcpy(buf,"%"); 
		break;
	case 'a': /* art */
		if (akt_brett->art==BINAER_BRETT)
			*buf='B';
		else
			*buf='T';
		buf[1]=0;
		break;
	case 'A': /* Sysopinfos */
		if (!akt_brett->aliasname)
			strcpy(buf,"");
		else
			strcpy(buf,akt_brett->aliasname);
		break;
	case 'c':                     /* c -> chef */
		sprintf(buf,"%lu",(ULONG) akt_brett->grp_chef);
		break;
	case 'd': /* last write access */
		time_t_to_mtzeit(&akt_brett->datum,buf);
		break;
	case 'D': /* created date */
		time_t_to_mtzeit(&akt_brett->created,buf);
		break;
	case 'e': /* letzte nderung an den Netzgruppendaten */
		time_t_to_mtzeit(&akt_brett->netzaenderung,buf);
		break;
	case 'E': /* letzte nderung an der Gruppenkonfiguration */
		time_t_to_mtzeit(&akt_brett->configaenderung,buf);
		break;
	case 'f': /* followup-to-default */
		if (akt_brett->followup_to_default)
			strcpy(buf,br_ind[akt_brett->followup_to_default].brett->name);
		else
			*buf=0;
		break;
	case 'i':                     /* i -> brett.index */
		sprintf(buf,"%lu",(ULONG) akt_brett->index);
		break;
	case 'L': /* Lokalinfos */
		if (!akt_brett->lokalinfo)
			strcpy(buf,"");
		else
			strcpy(buf,akt_brett->lokalinfo);
		break;
	case 'N':                     /* N -> brett.name */
		if (akt_brett->nr)
			strcpy(buf,akt_brett->name);
		else
			strcpy(buf,"Postfach");
		break;
	case 'n':                     /* n -> brett.nummer */
		sprintf(buf,"%lu",(ULONG) akt_brett->nr);
		break;
	case 'p':                     /* partition */
		*buf=akt_brett->partition + 'A';
		buf[1]=0;
	case 'P':                     /* Passwort */
		if (akt_brett->passwort && *akt_brett->passwort)
			strcpy(buf,akt_brett->passwort);
		else
			*buf=0;
		break;
	case 'r':                     /* r -> read-level (lesen_level) */
		sprintf(buf,"%lu",(ULONG) (akt_brett->nr? akt_brett->lesen_level : 0));
		break;
	case 'w':                     /* w -> write-level (schreib_level) */
		sprintf(buf,"%lu",(ULONG) (akt_brett->nr ? akt_brett->schreib_level : 0));
		break;
	case 'S':                     /* S -> status */
		if (akt_brett->nr)
			brettflags_to_string(akt_brett->flags, buf, sizeof(buf));
		else
			*buf=0;
		break;
	case 's': /* sprache */
		{
			const char *sp=bb_sprache(akt_brett->sprache);
			if (sp)
				strcpy(buf,sp);
			else
				strcpy(buf,"Unbekannt (?)");
		}
		break;
	case 'u': /* ursprungsnetz */
		{
			const char *sp=bb_ursprungsnetz(akt_brett->ursprungsnetz);
			if (sp)
				strcpy(buf,sp);
			else
				strcpy(buf,"Unbekannt (?)");
		}
		break;
	case 'U': /* Userinfos */
		if (!akt_brett->userinfo)
			strcpy(buf,"");
		else
			strcpy(buf,akt_brett->userinfo);
		break;
	case 'V':                     /* V -> Vernetzt*/
		strcpy(buf,"N");
		if (akt_brett->nr && akt_brett->netzstatus)
			*buf='J';
		break;
	case 'Y': /* Sysopinfos */
		if (!akt_brett->sysopinfo)
			strcpy(buf,"");
		else
			strcpy(buf,akt_brett->sysopinfo);
		break;
	default:
		return TIP_NOTDONE;
	}

	l=strlen(buf);
	if (tmaxlen!=-1 && l>tmaxlen)
	{
		l=tmaxlen;
		buf[tmaxlen]=0;
	}
	if (l>=maxlen || (tfixlen!=-1 && tfixlen>=maxlen))
	{
		tip_errtext="buffer shortage"; 
		return TIP_ERROR;
	}
	if (tfixlen!=-1)
	{
		retl=tfixlen;
		while (tfixlen>l)
		{
			tfixlen--;
			ADDC(pad);
		}
	}
	else
		retl=l;
	strcpy(to,buf);
	return retl;
}

static BOOLEAN
do_interpret(const char *format, char *to, size_t max)
{
	*to=0;
	if (tip_interpret(tiptab,format,to,max)!=TIP_ERROR)
		return TRUE;
	fprintf(stderr,"Fehler bei Textinterpretation: %s\n",tip_errtext);
	return FALSE;
}

static void
usage(int ecode)
{
	fprintf(stderr,
		"usage: %s [-d db-dir] [-s system-dir] [-f format] [-l] [-r]\n", 
		program_name);
	exit(ecode);
}

int 
main(int argc, char **argv)
{
	const char * tom_var_dir="/usr/local/quark/var/";
	const char * tom_text_db_dir="/usr/local/quark/bretter/";
	const char * tom_bin_db_dir="/usr/local/quark/binbretter/";
	const char * format="%n %N\n";
	const char * wenn=NULL;
	const char * regular=NULL;
	const char * wc=NULL;
	enum { LISTMODE, REORGMODE} mode=LISTMODE;
	BRETTNR *srt;
	struct option opts[] =
	{
		{"bin", required_argument, NULL, 'd'},
		{"format", required_argument, NULL, 'f'},
		{"if", required_argument, NULL, 'i'},
		{"list", no_argument,NULL, 'l'},
		{"reorg", no_argument, NULL, 'r'},
		{"regex", required_argument, NULL, 'R'},
		{"text", required_argument, NULL, 't'},
		{"var", required_argument, NULL, 'v'},
		{"wildcard", required_argument, NULL, 'w'},
		{NULL, 0, NULL, 0}
	};
	int option_index = 0;

	program_name=strrchr(argv[0],'/');
	if (!program_name)
		program_name = argv[0];
	else
		program_name++;

	parse_long_options (argc, argv, program_name, PACKAGE_VERSION, usage);

	while (1)
	{
		int c;

		c = getopt_long (argc, argv, "d:f:i:lrR:t:v:w:", opts, &option_index);

        if (c == -1)
            break;
        switch (c)
        {
        case 'b':
        	if (optarg[strlen(optarg)-1]=='/')
				tom_bin_db_dir=XSTRDUP(optarg);
			else
			{
				tom_bin_db_dir=concat(optarg,"/",NULL);
				XNULLCHECK(tom_bin_db_dir);
			}
			break;
        case 'f': format=XSTRDUP(optarg); break;
		case 'i': wenn=XSTRDUP(optarg); break;
		case 'l': mode=LISTMODE; break;
		case 'r': mode=REORGMODE; break;
		case 'R': regular=XSTRDUP(optarg); break;
        case 't':
        	if (optarg[strlen(optarg)-1]=='/')
				tom_text_db_dir=XSTRDUP(optarg);
			else
			{
				tom_text_db_dir=concat(optarg,"/",NULL);
				XNULLCHECK(tom_text_db_dir);
			}
			break;
        case 'v':
        	if (optarg[strlen(optarg)-1]=='/')
				tom_var_dir=XSTRDUP(optarg);
			else
			{
				tom_var_dir=concat(optarg,"/",NULL);
				XNULLCHECK(tom_var_dir);
			}
			break;
		case 'w': wc=XSTRDUP(optarg); break;
		default:
			usage(2);
		}
	}

	if (!bb_init(tom_var_dir,tom_text_db_dir,tom_bin_db_dir))
		error(1,0,"cannot initialize database");

	srt=bb_sortierung();
	XNULLCHECK(srt);
	if (mode==LISTMODE)
	{
		int i;
		BRETTNR bnr;
		regex_t rx;

		if (regular)
		{
			if (regcomp(&rx,regular,REG_ICASE | REG_NOSUB)!=0)
				error(1,0,"can't compile %s",regular);
		}
		for (i=0;srt[i];i++)
		{
			char buf[4096];
			bnr=srt[i];
			if (!br_ind[bnr].brett)
				continue;
			/* aus dem Cache holen */
			akt_brett=br_ind[bnr].brett;

			if (wenn)
			{
				if (!do_interpret(wenn,buf,sizeof(buf)))
					error(1,0,"can't interpret %s for %s",
							wenn,akt_brett->name);
				/* Kleine Optimierung */
				if (!wc && (!regular || 0==strcmp(regular,".")))
				{
					if (!*buf)
						continue;
				}
				else if (!wc)
				{
					/* regulrer Ausdruck */
					if (0!=regexec(&rx,buf,0,NULL,0))
						continue;
				}
				else 
				{
					/* wildcard */
#ifdef FNM_CASEFOLD
					if (0!=fnmatch(wc,buf,FNM_CASEFOLD))
#else
					char buf1[4096];
					char buf2[4096];
					entumlauten(wc,buf1);
					entumlauten(buf,buf2);
					strlwr(buf1);
					strlwr(buf2);
					if (0!=fnmatch(buf1,buf2,0))
#endif
						continue;
				}
			}
			if (!do_interpret(format,buf,sizeof(buf)))
				error(1,0,"can't interpret %s for %s",
							format,akt_brett->name);
			fputs(buf,stdout);
		}
	}
	else if (mode==REORGMODE)
	{
		if (!bb_reorg())
			error(1,0,"reorganization failed");
	}

#ifdef LEAK_CHECK
	GC_gcollect();
#endif
	exit(0);
}
